import { 
    world, 
    Player,
    EntityDamageCause,
    system
} from "@minecraft/server";

// Totem Counter Mod
// Created By_IIBl4z3MasterII

class TotemCounter {
    static #instance = null;
    #lastDamageSource = new Map();
    #lastUsedMessages = new Set();
    
    static get instance() {
        if (!TotemCounter.#instance) {
            TotemCounter.#instance = new TotemCounter();
        }
        return TotemCounter.#instance;
    }

    constructor() {
        if (TotemCounter.#instance) {
            return TotemCounter.#instance;
        }

        this.#initializeConstants();
        this.#initializeWorld();
        this.#setupEvents();
    }

    #initializeConstants() {
        this.CONSTANTS = Object.freeze({
            CLEANUP_DELAY: 100,
            DAMAGE_WINDOW: 1000,
            MAX_STORED_MESSAGES: 5,
            PROPERTIES: {
                GLOBAL_COUNT: "globalTotemCount_By_IIBl4z3MasterII",
                PLAYER_COUNT: "totemUsageCount_By_IIBl4z3MasterII"
            },
            MESSAGE: {
                PREFIX: "§7[§6Tótem§7] ",
                SUFFIX: count => `\n§7Total del servidor: §e${count} §7tótems usados`
            }
        });
    }

    #initializeWorld() {
        const globalCount = world.getDynamicProperty(this.CONSTANTS.PROPERTIES.GLOBAL_COUNT);
        if (globalCount === undefined) {
            world.setDynamicProperty(this.CONSTANTS.PROPERTIES.GLOBAL_COUNT, 0);
        }
    }

    #getDamageCauseText(cause) {
        const causes = {
            [EntityDamageCause.anvil]: "aplastamiento por yunque",
            [EntityDamageCause.blockExplosion]: "una poderosa explosión",
            [EntityDamageCause.charging]: "una brutal embestida",
            [EntityDamageCause.contact]: "contacto con bloques dañinos",
            [EntityDamageCause.drowning]: "ahogamiento en las profundidades",
            [EntityDamageCause.entityAttack]: "un ataque mortal",
            [EntityDamageCause.entityExplosion]: "una devastadora explosión",
            [EntityDamageCause.fall]: "una caída mortal",
            [EntityDamageCause.fallingBlock]: "aplastamiento por bloques",
            [EntityDamageCause.fire]: "las llamas ardientes",
            [EntityDamageCause.fireworks]: "una explosión de fuegos artificiales",
            [EntityDamageCause.flyIntoWall]: "un impacto brutal contra la pared",
            [EntityDamageCause.freezing]: "congelamiento extremo",
            [EntityDamageCause.lava]: "un baño en lava",
            [EntityDamageCause.lightning]: "un poderoso rayo",
            [EntityDamageCause.magic]: "magia oscura",
            [EntityDamageCause.magma]: "el ardiente magma",
            [EntityDamageCause.override]: "una fuerza sobrenatural",
            [EntityDamageCause.piston]: "la fuerza de un pistón",
            [EntityDamageCause.projectile]: "un proyectil letal",
            [EntityDamageCause.stalactite]: "una afilada estalactita",
            [EntityDamageCause.stalagmite]: "una puntiaguda estalagmita",
            [EntityDamageCause.starve]: "la inanición",
            [EntityDamageCause.suffocation]: "la asfixia",
            [EntityDamageCause.suicide]: "un intento de suicidio",
            [EntityDamageCause.temperature]: "temperaturas extremas",
            [EntityDamageCause.thorns]: "espinas venenosas",
            [EntityDamageCause.void]: "el vacío infinito",
            [EntityDamageCause.wither]: "el efecto del wither"
        };
        return causes[cause] ?? "poderes misteriosos";
    }

    #getRandomMessage(playerName, cause, count, globalCount) {
        const messages = [
            `§e${playerName} §fha desafiado a la muerte por §c${cause} §fy ha sobrevivido §e(${count} tótems)`,
            `§e${playerName} §fha escapado de §c${cause} §fgracias al poder del tótem §e(${count} veces)`,
            `§e${playerName} §fha burlado a la muerte causada por §c${cause} §e(${count} tótems usados)`,
            `§e${playerName} §fha resistido §c${cause} §fcon la ayuda de un tótem §e(Total: ${count})`,
            `§fEl tótem ha salvado a §e${playerName} §fde §c${cause} §e(${count} resurrecciones)`,
            `§e${playerName} §fha sobrevivido milagrosamente a §c${cause} §e(${count} tótems consumidos)`,
            `§fLa magia ancestral protegió a §e${playerName} §fde §c${cause} §e(${count} tótems)`,
            `§e${playerName} §fha engañado a la muerte provocada por §c${cause} §e(${count} veces)`,
            `§fUn tótem más consumido por §e${playerName} §fpara salvarse de §c${cause} §e(${count} total)`,
            `§fLos dioses antiguos protegieron a §e${playerName} §fde §c${cause} §e(${count} tótems)`,
            `§e${playerName} §fha sido bendecido con otra vida tras §c${cause} §e(${count} resurrecciones)`,
            `§fLa muerte intentó llevarse a §e${playerName} §fpor §c${cause}§f, pero fracasó §e(${count} veces)`,
            `§e${playerName} §fha sido salvado por la magia del tótem de §c${cause} §e(${count} salvaciones)`,
            `§fEl poder ancestral ayudó a §e${playerName} §fa sobrevivir §c${cause} §e(${count} tótems)`,
            `§e${playerName} §fha recibido el favor de los tótems contra §c${cause} §e(${count} bendiciones)`,
            `§fLa muerte tendrá que esperar para llevarse a §e${playerName} §ftras §c${cause} §e(${count} tótems)`,
            `§e${playerName} §fha sido rescatado de §c${cause} §fpor la magia antigua §e(${count} rescates)`,
            `§fLos espíritus del tótem salvaron a §e${playerName} §fde §c${cause} §e(${count} veces)`,
            `§e${playerName} §fha desafiado el destino contra §c${cause} §e(${count} desafíos superados)`,
            `§fLa muerte no pudo reclamar a §e${playerName} §ftras §c${cause} §e(${count} victorias)`
        ];

        let availableMessages = messages.filter(msg => !this.#lastUsedMessages.has(msg));
        
        if (availableMessages.length === 0) {
            this.#lastUsedMessages.clear();
            availableMessages = messages;
        }

        const selectedMessage = availableMessages[Math.floor(Math.random() * availableMessages.length)];
        
        this.#updateLastUsedMessages(selectedMessage);

        return this.CONSTANTS.MESSAGE.PREFIX + 
               selectedMessage + 
               this.CONSTANTS.MESSAGE.SUFFIX(globalCount);
    }

    #updateLastUsedMessages(message) {
        this.#lastUsedMessages.add(message);
        if (this.#lastUsedMessages.size > this.CONSTANTS.MAX_STORED_MESSAGES) {
            this.#lastUsedMessages.delete(this.#lastUsedMessages.values().next().value);
        }
    }

    #cleanupPlayerData(playerId) {
        system.runTimeout(() => {
            this.#lastDamageSource.delete(playerId);
        }, this.CONSTANTS.CLEANUP_DELAY);
    }

    #handlePlayerDamage = (event) => {
        if (!(event.hurtEntity instanceof Player)) return;
        
        const player = event.hurtEntity;
        this.#lastDamageSource.set(player.id, {
            cause: event.damageSource.cause,
            amount: event.damage,
            time: Date.now()
        });
    }

    #handleHealthChange = (event) => {
        if (!(event.entity instanceof Player)) return;
        
        const player = event.entity;
        if (event.oldValue <= 0 && event.newValue > 0) {
            this.#processTotemUse(player);
        }
    }

    #processTotemUse(player) {
        const damageInfo = this.#lastDamageSource.get(player.id);
        if (!damageInfo || Date.now() - damageInfo.time > this.CONSTANTS.DAMAGE_WINDOW) return;

        const playerCount = this.#updatePlayerCount(player);
        const globalCount = this.#updateGlobalCount();

        const cause = this.#getDamageCauseText(damageInfo.cause);
        const message = this.#getRandomMessage(player.name, cause, playerCount, globalCount);
        world.sendMessage(message);

        this.#cleanupPlayerData(player.id);
    }

    #updatePlayerCount(player) {
        const currentCount = player.getDynamicProperty(this.CONSTANTS.PROPERTIES.PLAYER_COUNT) ?? 0;
        const newCount = currentCount + 1;
        player.setDynamicProperty(this.CONSTANTS.PROPERTIES.PLAYER_COUNT, newCount);
        return newCount;
    }

    #updateGlobalCount() {
        const currentCount = world.getDynamicProperty(this.CONSTANTS.PROPERTIES.GLOBAL_COUNT) ?? 0;
        const newCount = currentCount + 1;
        world.setDynamicProperty(this.CONSTANTS.PROPERTIES.GLOBAL_COUNT, newCount);
        return newCount;
    }

    #setupEvents() {
        world.afterEvents.entityHurt.subscribe(this.#handlePlayerDamage);
        world.afterEvents.entityHealthChanged.subscribe(this.#handleHealthChange);
    }
}

// Exportamos una instancia única usando el getter estático
export const totemCounter = TotemCounter.instance;